import asyncio
import json
import random
import logging
import hashlib
import inspect
from aiohttp import web, hdrs

from sockjs.session import SessionManager
from sockjs.protocol import IFRAME_HTML
from sockjs.transports import handlers
from sockjs.transports.utils import CACHE_CONTROL
from sockjs.transports.utils import session_cookie
from sockjs.transports.utils import cors_headers
from sockjs.transports.utils import cache_headers
from sockjs.transports.rawwebsocket import RawWebSocketTransport

log = logging.getLogger("sockjs")


def get_manager(name, app):
    return app["__sockjs_managers__"][name]


def _gen_endpoint_name():
    return "n" + str(random.randint(1000, 9999))


def add_endpoint(
    app,
    handler,
    *,
    name="",
    prefix="/sockjs",
    manager=None,
    disable_transports=(),
    sockjs_cdn="https://cdn.jsdelivr.net/npm/sockjs-client@1/dist/sockjs.min.js",  # noqa
    cookie_needed=True
):

    assert callable(handler), handler
    if not asyncio.iscoroutinefunction(handler) and not inspect.isgeneratorfunction(
        handler
    ):
        handler = asyncio.coroutine(handler)

    router = app.router

    if not name:
        name = _gen_endpoint_name()

    # set session manager
    if manager is None:
        manager = SessionManager(name, app, handler)

    if manager.name != name:
        raise ValueError("Session manage must have same name as sockjs route")

    managers = app.setdefault("__sockjs_managers__", {})
    if name in managers:
        raise ValueError('SockJS "%s" route already registered' % name)

    managers[name] = manager

    # register routes
    route = SockJSRoute(
        name, manager, sockjs_cdn, handlers, disable_transports, cookie_needed
    )

    if prefix.endswith("/"):
        prefix = prefix[:-1]

    route_name = "sockjs-url-%s-greeting" % name
    router.add_route(hdrs.METH_GET, prefix, route.greeting, name=route_name)

    route_name = "sockjs-url-%s" % name
    router.add_route(hdrs.METH_GET, "%s/" % prefix, route.greeting, name=route_name)

    route_name = "sockjs-%s" % name
    router.add_route(
        hdrs.METH_ANY,
        "%s/{server}/{session}/{transport}" % prefix,
        route.handler,
        name=route_name,
    )

    route_name = "sockjs-websocket-%s" % name
    router.add_route(
        hdrs.METH_GET, "%s/websocket" % prefix, route.websocket, name=route_name
    )

    router.add_route(
        hdrs.METH_GET, "%s/info" % prefix, route.info, name="sockjs-info-%s" % name
    )
    router.add_route(
        hdrs.METH_OPTIONS,
        "%s/info" % prefix,
        route.info_options,
        name="sockjs-info-options-%s" % name,
    )

    route_name = "sockjs-iframe-%s" % name
    router.add_route(
        hdrs.METH_GET, "%s/iframe.html" % prefix, route.iframe, name=route_name
    )

    route_name = "sockjs-iframe-ver-%s" % name
    router.add_route(
        hdrs.METH_GET, "%s/iframe{version}.html" % prefix, route.iframe, name=route_name
    )

    # start session gc
    manager.start()


class SockJSRoute:
    def __init__(
        self,
        name,
        manager,
        sockjs_cdn,
        handlers,
        disable_transports,
        cookie_needed=True,
    ):
        self.name = name
        self.manager = manager
        self.handlers = handlers
        self.disable_transports = dict((k, 1) for k in disable_transports)
        self.cookie_needed = cookie_needed
        self.iframe_html = (IFRAME_HTML % sockjs_cdn).encode("utf-8")
        self.iframe_html_hxd = hashlib.md5(self.iframe_html).hexdigest()

    async def handler(self, request):
        info = request.match_info

        # lookup transport
        tid = info["transport"]

        if tid not in self.handlers or tid in self.disable_transports:
            return web.HTTPNotFound()

        create, transport = self.handlers[tid]

        # session
        manager = self.manager
        if not manager.started:
            manager.start()

        sid = info["session"]
        if not sid or "." in sid or "." in info["server"]:
            return web.HTTPNotFound()

        try:
            session = manager.get(sid, create, request=request)
        except KeyError:
            return web.HTTPNotFound(headers=session_cookie(request))

        t = transport(manager, session, request)
        try:
            return await t.process()
        except asyncio.CancelledError:
            raise
        except web.HTTPException as exc:
            return exc
        except Exception:
            log.exception("Exception in transport: %s" % tid)
            if manager.is_acquired(session):
                await manager.release(session)
            return web.HTTPInternalServerError()

    async def websocket(self, request):
        # session
        sid = "%0.9d" % random.randint(1, 2147483647)
        session = self.manager.get(sid, True, request=request)

        transport = RawWebSocketTransport(self.manager, session, request)
        try:
            return await transport.process()
        except asyncio.CancelledError:
            raise
        except web.HTTPException as exc:
            return exc

    async def info(self, request):
        resp = web.Response()
        resp.headers[hdrs.CONTENT_TYPE] = "application/json;charset=UTF-8"
        resp.headers[hdrs.CACHE_CONTROL] = CACHE_CONTROL
        resp.headers.extend(cors_headers(request.headers))

        info = {
            "entropy": random.randint(1, 2147483647),
            "websocket": "websocket" not in self.disable_transports,
            "cookie_needed": self.cookie_needed,
            "origins": ["*:*"],
        }
        resp.text = json.dumps(info)
        return resp

    async def info_options(self, request):
        resp = web.Response(status=204)
        resp.headers[hdrs.CONTENT_TYPE] = "application/json;charset=UTF-8"
        resp.headers[hdrs.CACHE_CONTROL] = CACHE_CONTROL
        resp.headers[hdrs.ACCESS_CONTROL_ALLOW_METHODS] = "OPTIONS, GET"
        resp.headers.extend(cors_headers(request.headers))
        resp.headers.extend(cache_headers())
        resp.headers.extend(session_cookie(request))
        return resp

    async def iframe(self, request):
        cached = request.headers.get(hdrs.IF_NONE_MATCH)
        if cached:
            response = web.Response(status=304)
            response.headers[hdrs.CONTENT_TYPE] = ""
            response.headers.extend(cache_headers())
            return response

        headers = (
            (hdrs.CONTENT_TYPE, "text/html;charset=UTF-8"),
            (hdrs.ETAG, self.iframe_html_hxd),
        )
        headers += cache_headers()
        return web.Response(body=self.iframe_html, headers=headers)

    async def greeting(self, request):
        return web.Response(
            body=b"Welcome to SockJS!\n",
            headers=((hdrs.CONTENT_TYPE, "text/plain; charset=UTF-8"),),
        )



""" iframe-eventsource transport """
from aiohttp import web, hdrs
from sockjs.protocol import ENCODING

from .base import StreamingTransport
from .utils import CACHE_CONTROL, session_cookie


class EventsourceTransport(StreamingTransport):
    async def send(self, text):
        blob = "".join(("data: ", text, "\r\n\r\n")).encode(ENCODING)
        await self.response.write(blob)

        self.size += len(blob)
        if self.size > self.maxsize:
            return True
        else:
            return False

    async def process(self):
        headers = (
            (hdrs.CONTENT_TYPE, "text/event-stream"),
            (hdrs.CACHE_CONTROL, CACHE_CONTROL),
        )
        headers += session_cookie(self.request)

        # open sequence (sockjs protocol)
        resp = self.response = web.StreamResponse(headers=headers)
        await resp.prepare(self.request)
        await resp.write(b"\r\n")

        # handle session
        await self.handle_session()

        return resp



import collections
import hashlib
from datetime import datetime

ENCODING = "utf-8"

STATE_NEW = 0
STATE_OPEN = 1
STATE_CLOSING = 2
STATE_CLOSED = 3


_days = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
_months = [
    "Jan",
    "Feb",
    "Mar",
    "Apr",
    "May",
    "Jun",
    "Jul",
    "Aug",
    "Sep",
    "Oct",
    "Nov",
    "Dec",
]

# json
# -----------

# Fastest
try:
    import ujson as json

    kwargs = {}  # pragma: no cover
except ImportError:  # pragma: no cover

    def dthandler(obj):
        if isinstance(obj, datetime):
            now = obj.timetuple()
            return "%s, %02d %s %04d %02d:%02d:%02d -0000" % (
                _days[now[6]],
                now[2],
                _months[now[1] - 1],
                now[0],
                now[3],
                now[4],
                now[5],
            )

    kwargs = {"default": dthandler, "separators": (",", ":")}

    # Faster
    try:
        import simplejson as json
    except ImportError:
        # Slowest
        import json


# Frames
# ------

FRAME_OPEN = "o"
FRAME_CLOSE = "c"
FRAME_MESSAGE = "a"
FRAME_MESSAGE_BLOB = "a1"
FRAME_HEARTBEAT = "h"


# ------------------

IFRAME_HTML = """<!DOCTYPE html>
<html>
<head>
<meta http-equiv="X-UA-Compatible" content="IE=edge" />
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
  <script src="%s"></script>
  <script>
    document.domain = document.domain;
    SockJS.bootstrap_iframe();
  </script>
</head>
<body>
  <h2>Don't panic!</h2>
  <p>This is a SockJS hidden iframe. It's used for cross domain magic.</p>
</body>
</html>
""".strip()


IFRAME_MD5 = hashlib.md5(IFRAME_HTML.encode()).hexdigest()

loads = json.loads
ENCODING = "utf-8"


def dumps(data):
    return json.dumps(data, **kwargs)


def close_frame(code, reason):
    return FRAME_CLOSE + json.dumps([code, reason], **kwargs)


def message_frame(message):
    return FRAME_MESSAGE + json.dumps([message], **kwargs)


def messages_frame(messages):
    return FRAME_MESSAGE + json.dumps(messages, **kwargs)


# Handler messages
# ---------------------

MSG_OPEN = 1
MSG_MESSAGE = 2
MSG_CLOSE = 3
MSG_CLOSED = 4


class SockjsMessage(collections.namedtuple("SockjsMessage", ["type", "data"])):
    @property
    def tp(self):
        return self.type


OpenMessage = SockjsMessage(MSG_OPEN, None)
CloseMessage = SockjsMessage(MSG_CLOSE, None)
ClosedMessage = SockjsMessage(MSG_CLOSED, None)



import asyncio
from unittest import mock
from datetime import datetime, timedelta

from aiohttp import web

import pytest

from asyncio import ensure_future

from sockjs import Session, SessionIsClosed, protocol, SessionIsAcquired


class TestSession:
    async def test_ctor(self, mocker, make_handler, make_request):
        dt = mocker.patch("sockjs.session.datetime")
        now = dt.now.return_value = datetime.now()

        handler = make_handler([])
        request = make_request("GET", "/TestPath")
        session = Session("id", handler, request)

        assert session.id == "id"
        assert not session.expired
        assert session.expires == now + timedelta(seconds=10)

        assert session._hits == 0
        assert session._heartbeats == 0
        assert session.state == protocol.STATE_NEW

        session = Session("id", handler, request, timeout=timedelta(seconds=15))

        assert session.id == "id"
        assert not session.expired
        assert session.expires == now + timedelta(seconds=15)

    async def test_str(self, make_session):
        session = make_session("test")
        session.state = protocol.STATE_OPEN

        assert str(session) == "id='test' connected"

        session._hits = 10
        session._heartbeats = 50
        session.state = protocol.STATE_CLOSING
        assert str(session) == "id='test' disconnected hits=10 heartbeats=50"

        session._feed(protocol.FRAME_MESSAGE, "msg")
        assert str(session) == "id='test' disconnected queue[1] hits=10 heartbeats=50"

        session.state = protocol.STATE_CLOSED
        assert str(session) == "id='test' closed queue[1] hits=10 heartbeats=50"

        session.state = protocol.STATE_OPEN
        session.acquired = True
        expected = "id='test' connected acquired queue[1] hits=10 heartbeats=50"
        assert str(session) == expected

    async def test_tick(self, mocker, make_session):
        dt = mocker.patch("sockjs.session.datetime")
        now = dt.now.return_value = datetime.now()
        session = make_session("test")

        now = dt.now.return_value = now + timedelta(hours=1)
        session._tick()
        assert session.expires == now + session.timeout

    async def test_tick_different_timeoutk(self, mocker, make_session):
        dt = mocker.patch("sockjs.session.datetime")
        now = dt.now.return_value = datetime.now()
        session = make_session("test", timeout=timedelta(seconds=20))

        now = dt.now.return_value = now + timedelta(hours=1)
        session._tick()
        assert session.expires == now + timedelta(seconds=20)

    async def test_tick_custom(self, mocker, make_session):
        dt = mocker.patch("sockjs.session.datetime")
        now = dt.now.return_value = datetime.now()
        session = make_session("test", timeout=timedelta(seconds=20))

        now = dt.now.return_value = now + timedelta(hours=1)
        session._tick(timedelta(seconds=30))
        assert session.expires == now + timedelta(seconds=30)

    async def test_heartbeat(self, make_session):
        session = make_session("test")
        assert session._heartbeats == 0
        session._heartbeat()
        assert session._heartbeats == 1
        session._heartbeat()
        assert session._heartbeats == 2

    async def test_heartbeat_transport(self, make_session):
        session = make_session("test")
        session._heartbeat_transport = True
        session._heartbeat()
        assert list(session._queue) == [
            (protocol.FRAME_HEARTBEAT, protocol.FRAME_HEARTBEAT)
        ]

    async def test_expire(self, make_session):
        session = make_session("test")
        assert not session.expired

        session.expire()
        assert session.expired

    async def test_send(self, make_session):
        session = make_session("test")
        session.send("message")
        assert list(session._queue) == []

        session.state = protocol.STATE_OPEN
        session.send("message")

        assert list(session._queue) == [(protocol.FRAME_MESSAGE, ["message"])]

    async def test_send_non_str(self, make_session):
        session = make_session("test")
        with pytest.raises(AssertionError):
            session.send(b"str")

    async def test_send_frame(self, make_session):
        session = make_session("test")
        session.send_frame('a["message"]')
        assert list(session._queue) == []

        session.state = protocol.STATE_OPEN
        session.send_frame('a["message"]')

        assert list(session._queue) == [(protocol.FRAME_MESSAGE_BLOB, 'a["message"]')]

    async def test_feed(self, make_session):
        session = make_session("test")
        session._feed(protocol.FRAME_OPEN, protocol.FRAME_OPEN)
        session._feed(protocol.FRAME_MESSAGE, "msg")
        session._feed(protocol.FRAME_CLOSE, (3001, "reason"))

        assert list(session._queue) == [
            (protocol.FRAME_OPEN, protocol.FRAME_OPEN),
            (protocol.FRAME_MESSAGE, ["msg"]),
            (protocol.FRAME_CLOSE, (3001, "reason")),
        ]

    async def test_feed_msg_packing(self, make_session):
        session = make_session("test")
        session._feed(protocol.FRAME_MESSAGE, "msg1")
        session._feed(protocol.FRAME_MESSAGE, "msg2")
        session._feed(protocol.FRAME_CLOSE, (3001, "reason"))
        session._feed(protocol.FRAME_MESSAGE, "msg3")

        assert list(session._queue) == [
            (protocol.FRAME_MESSAGE, ["msg1", "msg2"]),
            (protocol.FRAME_CLOSE, (3001, "reason")),
            (protocol.FRAME_MESSAGE, ["msg3"]),
        ]

    async def test_feed_with_waiter(self, make_session):
        loop = asyncio.get_event_loop()
        session = make_session("test")
        session._waiter = waiter = loop.create_future()
        session._feed(protocol.FRAME_MESSAGE, "msg")

        assert list(session._queue) == [(protocol.FRAME_MESSAGE, ["msg"])]
        assert session._waiter is None
        assert waiter.done()

    async def test_wait(self, make_session):
        s = make_session("test")
        s.state = protocol.STATE_OPEN

        async def send():
            await asyncio.sleep(0.001)
            s._feed(protocol.FRAME_MESSAGE, "msg1")

        ensure_future(send())
        frame, payload = await s._wait()
        assert frame == protocol.FRAME_MESSAGE
        assert payload == 'a["msg1"]'

    async def test_wait_closed(self, make_session):
        s = make_session("test")
        s.state = protocol.STATE_CLOSED
        with pytest.raises(SessionIsClosed):
            await s._wait()

    async def test_wait_message(self, make_session):
        s = make_session("test")
        s.state = protocol.STATE_OPEN
        s._feed(protocol.FRAME_MESSAGE, "msg1")
        frame, payload = await s._wait()
        assert frame == protocol.FRAME_MESSAGE
        assert payload == 'a["msg1"]'

    async def test_wait_close(self, make_session):
        s = make_session("test")
        s.state = protocol.STATE_OPEN
        s._feed(protocol.FRAME_CLOSE, (3000, "Go away!"))
        frame, payload = await s._wait()
        assert frame == protocol.FRAME_CLOSE
        assert payload == 'c[3000,"Go away!"]'

    async def test_wait_message_unpack(self, make_session):
        s = make_session("test")
        s.state = protocol.STATE_OPEN
        s._feed(protocol.FRAME_MESSAGE, "msg1")
        frame, payload = await s._wait(pack=False)
        assert frame == protocol.FRAME_MESSAGE
        assert payload == ["msg1"]

    async def test_wait_close_unpack(self, make_session):
        s = make_session("test")
        s.state = protocol.STATE_OPEN
        s._feed(protocol.FRAME_CLOSE, (3000, "Go away!"))
        frame, payload = await s._wait(pack=False)
        assert frame == protocol.FRAME_CLOSE
        assert payload == (3000, "Go away!")

    async def test_close(self, make_session):
        session = make_session("test")
        session.state = protocol.STATE_OPEN
        session.close()
        assert session.state == protocol.STATE_CLOSING
        assert list(session._queue) == [(protocol.FRAME_CLOSE, (3000, "Go away!"))]

    async def test_close_idempotent(self, make_session):
        session = make_session("test")
        session.state = protocol.STATE_CLOSED
        session.close()
        assert session.state == protocol.STATE_CLOSED
        assert list(session._queue) == []

    async def test_acquire_new_session(self, make_session):
        manager = object()
        messages = []

        session = make_session(result=messages)
        assert session.state == protocol.STATE_NEW

        await session._acquire(manager)
        assert session.state == protocol.STATE_OPEN
        assert session.manager is manager
        assert session._heartbeat_transport
        assert list(session._queue) == [(protocol.FRAME_OPEN, protocol.FRAME_OPEN)]
        assert messages == [(protocol.OpenMessage, session)]

    async def test_acquire_exception_in_handler(self, make_session):
        async def handler(msg, s):
            raise ValueError

        session = make_session(handler=handler)
        assert session.state == protocol.STATE_NEW

        await session._acquire(object())
        assert session.state == protocol.STATE_CLOSING
        assert session._heartbeat_transport
        assert session.interrupted
        assert list(session._queue) == [
            (protocol.FRAME_OPEN, protocol.FRAME_OPEN),
            (protocol.FRAME_CLOSE, (3000, "Internal error")),
        ]

    async def test_remote_close(self, make_session):
        messages = []
        session = make_session(result=messages)

        await session._remote_close()
        assert not session.interrupted
        assert session.state == protocol.STATE_CLOSING
        assert messages == [(protocol.SockjsMessage(protocol.MSG_CLOSE, None), session)]

    async def test_remote_close_idempotent(self, make_session):
        messages = []
        session = make_session(result=messages)
        session.state = protocol.STATE_CLOSED

        await session._remote_close()
        assert session.state == protocol.STATE_CLOSED
        assert messages == []

    async def test_remote_close_with_exc(self, make_session):
        messages = []
        session = make_session(result=messages)

        exc = ValueError()
        await session._remote_close(exc=exc)
        assert session.interrupted
        assert session.state == protocol.STATE_CLOSING
        assert messages == [(protocol.SockjsMessage(protocol.MSG_CLOSE, exc), session)]

    async def test_remote_close_exc_in_handler(self, make_session, make_handler):
        handler = make_handler([], exc=True)
        session = make_session(handler=handler)

        await session._remote_close()
        assert not session.interrupted
        assert session.state == protocol.STATE_CLOSING

    async def test_remote_closed(self, make_session):
        messages = []
        session = make_session(result=messages)

        await session._remote_closed()
        assert session.expired
        assert session.state == protocol.STATE_CLOSED
        assert messages == [(protocol.ClosedMessage, session)]

    async def test_remote_closed_idempotent(self, make_session):
        messages = []
        session = make_session(result=messages)
        session.state = protocol.STATE_CLOSED

        await session._remote_closed()
        assert session.state == protocol.STATE_CLOSED
        assert messages == []

    async def test_remote_closed_with_waiter(self, make_session):
        messages = []
        session = make_session(result=messages)
        loop = asyncio.get_event_loop()
        session._waiter = waiter = loop.create_future()

        await session._remote_closed()
        assert waiter.done()
        assert session.expired
        assert session._waiter is None
        assert session.state == protocol.STATE_CLOSED
        assert messages == [(protocol.ClosedMessage, session)]

    async def test_remote_closed_exc_in_handler(self, make_handler, make_session):
        handler = make_handler([], exc=True)
        session = make_session(handler=handler)

        await session._remote_closed()
        assert session.expired
        assert session.state == protocol.STATE_CLOSED

    async def test_remote_message(self, make_session):
        messages = []
        session = make_session(result=messages)

        await session._remote_message("msg")
        assert messages == [
            (protocol.SockjsMessage(protocol.MSG_MESSAGE, "msg"), session)
        ]

    async def test_remote_message_exc(self, make_handler, make_session):
        messages = []
        handler = make_handler(messages, exc=True)
        session = make_session(handler=handler)

        await session._remote_message("msg")
        assert messages == []

    async def test_remote_messages(self, make_session):
        messages = []
        session = make_session(result=messages)

        await session._remote_messages(("msg1", "msg2"))
        assert messages == [
            (protocol.SockjsMessage(protocol.MSG_MESSAGE, "msg1"), session),
            (protocol.SockjsMessage(protocol.MSG_MESSAGE, "msg2"), session),
        ]

    async def test_remote_messages_exc(self, make_handler, make_session):
        messages = []
        handler = make_handler(messages, exc=True)
        session = make_session(handler=handler)

        await session._remote_messages(("msg1", "msg2"))
        assert messages == []


class TestSessionManager:
    async def test_request_available(self, make_manager, make_request):
        sm = make_manager()
        s = sm.get("test", True, make_request("GET", "/test/"))
        assert isinstance(s.request, web.Request)

    async def test_fresh(self, make_manager, make_session):
        sm = make_manager()
        s = make_session()
        sm._add(s)
        assert "test" in sm

    async def test_add(self, make_manager, make_session):
        sm = make_manager()
        s = make_session()

        sm._add(s)
        assert "test" in sm
        assert sm["test"] is s
        assert s.manager is sm

    async def test_add_expired(self, make_manager, make_session):
        sm = make_manager()
        s = make_session()
        s.expire()

        with pytest.raises(ValueError):
            sm._add(s)

    async def test_get(self, make_manager, make_session):
        sm = make_manager()
        s = make_session()
        with pytest.raises(KeyError):
            sm.get("test")

        sm._add(s)
        assert sm.get("test") is s

    async def test_get_unknown_with_default(self, make_manager):
        sm = make_manager()
        default = object()

        item = sm.get("id", default=default)
        assert item is default

    async def test_get_with_create(self, make_manager):
        sm = make_manager()

        s = sm.get("test", True)
        assert s.id in sm
        assert isinstance(s, Session)

    async def test_acquire(self, make_manager, make_session):
        sm = make_manager()
        s1 = make_session()
        sm._add(s1)
        s1._acquire = mock.Mock()
        loop = asyncio.get_event_loop()
        s1._acquire.return_value = loop.create_future()
        s1._acquire.return_value.set_result(1)

        s2 = await sm.acquire(s1)

        assert s1 is s2
        assert s1.id in sm.acquired
        assert sm.acquired[s1.id]
        assert sm.is_acquired(s1)
        assert s1._acquire.called

    async def test_acquire_unknown(self, make_manager, make_session):
        sm = make_manager()
        s = make_session()
        with pytest.raises(KeyError):
            await sm.acquire(s)

    async def test_acquire_locked(self, make_manager, make_session):
        sm = make_manager()
        s = make_session()
        sm._add(s)
        await sm.acquire(s)

        with pytest.raises(SessionIsAcquired):
            await sm.acquire(s)

    async def test_release(self, make_manager):
        sm = make_manager()
        s = sm.get("test", True)
        s._release = mock.Mock()

        await sm.acquire(s)
        await sm.release(s)

        assert "test" not in sm.acquired
        assert not sm.is_acquired(s)
        assert s._release.called

    async def test_active_sessions(self, make_manager):
        sm = make_manager()

        s1 = sm.get("test1", True)
        s2 = sm.get("test2", True)
        s2.expire()

        active = list(sm.active_sessions())
        assert len(active) == 1
        assert s1 in active

    async def test_broadcast(self, make_manager):
        sm = make_manager()

        s1 = sm.get("test1", True)
        s1.state = protocol.STATE_OPEN
        s2 = sm.get("test2", True)
        s2.state = protocol.STATE_OPEN
        sm.broadcast("msg")

        assert list(s1._queue) == [(protocol.FRAME_MESSAGE_BLOB, 'a["msg"]')]
        assert list(s2._queue) == [(protocol.FRAME_MESSAGE_BLOB, 'a["msg"]')]

    async def test_clear(self, make_manager):
        sm = make_manager()

        s1 = sm.get("s1", True)
        s1.state = protocol.STATE_OPEN
        s2 = sm.get("s2", True)
        s2.state = protocol.STATE_OPEN

        await sm.clear()

        assert not bool(sm)
        assert s1.expired
        assert s2.expired
        assert s1.state == protocol.STATE_CLOSED
        assert s2.state == protocol.STATE_CLOSED

    async def test_heartbeat(self, make_manager):
        sm = make_manager()
        assert not sm.started
        assert sm._hb_task is None

        sm.start()
        assert sm.started
        assert sm._hb_handle is not None

        sm._heartbeat()
        assert sm._hb_task is not None

        hb_task = sm._hb_task

        sm.stop()
        assert not sm.started
        assert sm._hb_handle is None
        assert sm._hb_task is None
        assert hb_task._must_cancel

    async def test_heartbeat_task(self, make_manager):
        sm = make_manager()
        sm._hb_task = mock.Mock()

        await sm._heartbeat_task()
        assert sm.started
        assert sm._hb_task is None

    async def test_gc_expire(self, make_manager, make_session):
        sm = make_manager()
        s = make_session()

        sm._add(s)
        await sm.acquire(s)
        await sm.release(s)

        s.expires = datetime.now() - timedelta(seconds=30)

        await sm._heartbeat_task()
        assert s.id not in sm
        assert s.expired
        assert s.state == protocol.STATE_CLOSED

    async def test_gc_expire_acquired(self, make_manager, make_session):
        sm = make_manager()
        s = make_session()
        sm._add(s)
        await sm.acquire(s)
        s.expires = datetime.now() - timedelta(seconds=30)
        await sm._heartbeat_task()

        assert s.id not in sm
        assert s.id not in sm.acquired
        assert s.expired
        assert s.state == protocol.STATE_CLOSED

    async def test_gc_one_expire(self, make_manager, make_session):
        sm = make_manager()
        s1 = make_session("id1")
        s2 = make_session("id2")

        sm._add(s1)
        sm._add(s2)
        await sm.acquire(s1)
        await sm.acquire(s2)
        await sm.release(s1)
        await sm.release(s2)

        s1.expires = datetime.now() - timedelta(seconds=30)

        await sm._heartbeat_task()
        assert s1.id not in sm
        assert s2.id in sm

    async def test_emits_warning_on_del(self, make_manager, make_session):

        sm = make_manager()
        s1 = make_session("id1")
        s2 = make_session("id2")

        sm._add(s1)
        sm._add(s2)

        with pytest.warns(RuntimeWarning) as warning:
            getattr(sm, "__del__")()
            msg = (
                "Unclosed sessions! Please call "
                "`await SessionManager.clear()` before del"
            )
            assert warning[0].message.args[0] == msg

    async def test_does_not_emits_warning_on_del_if_no_sessions(
        self, make_manager, make_session
    ):

        sm = make_manager()
        s1 = make_session("id1")
        s2 = make_session("id2")

        sm._add(s1)
        sm._add(s2)

        await sm.clear()
        getattr(sm, "__del__")()



import asyncio

from datetime import timedelta
from unittest import mock

import pytest

from aiohttp import web
from aiohttp.web_urldispatcher import UrlMappingMatchInfo
from aiohttp.test_utils import make_mocked_request, make_mocked_coro
from multidict import CIMultiDict
from yarl import URL

from sockjs import Session, SessionManager, transports
from sockjs.route import SockJSRoute


@pytest.fixture
def app():
    return web.Application()


@pytest.fixture
def make_fut():
    def maker(val, makemock=True):
        loop = asyncio.get_event_loop()
        fut = loop.create_future()
        fut.set_result(val)

        if makemock:
            m = mock.Mock()
            m.return_value = fut
            return m
        else:
            return fut

    return maker


@pytest.fixture
def make_handler():
    def maker(result, coro=True, exc=False):
        if result is None:
            result = []
        output = result

        def handler(msg, s):
            if exc:
                raise ValueError((msg, s))
            output.append((msg, s))

        if coro:
            return asyncio.coroutine(handler)
        else:
            return handler

    return maker


@pytest.fixture
def make_route(make_handler, app):
    def maker(handlers=transports.handlers):
        handler = make_handler([])
        sm = SessionManager("sm", app, handler)
        return SockJSRoute("sm", sm, "http:sockjs-cdn", handlers, (), True)

    return maker


@pytest.fixture
def make_request(app):
    def maker(method, path, query_params={}, headers=None, match_info=None):
        path = URL(path)
        if query_params:
            path = path.with_query(query_params)

        if headers is None:
            headers = CIMultiDict(
                {
                    "HOST": "server.example.com",
                    "UPGRADE": "websocket",
                    "CONNECTION": "Upgrade",
                    "SEC-WEBSOCKET-KEY": "dGhlIHNhbXBsZSBub25jZQ==",
                    "ORIGIN": "http://example.com",
                    "SEC-WEBSOCKET-PROTOCOL": "chat, superchat",
                    "SEC-WEBSOCKET-VERSION": "13",
                }
            )

        writer = mock.Mock()
        writer.write_headers = make_mocked_coro(None)
        writer.write = make_mocked_coro(None)
        writer.drain = make_mocked_coro(None)
        transport = mock.Mock()
        transport._drain_helper = make_mocked_coro()
        loop = asyncio.get_event_loop()
        ret = make_mocked_request(method, str(path), headers, writer=writer, loop=loop)

        if match_info is None:
            match_info = UrlMappingMatchInfo({}, mock.Mock())
            match_info.add_app(app)
        ret._match_info = match_info
        return ret

    return maker


@pytest.fixture
def make_session(make_handler, make_request):
    def maker(
        name="test", timeout=timedelta(10), request=None, handler=None, result=None
    ):
        if request is None:
            request = make_request("GET", "/TestPath/")

        if handler is None:
            handler = make_handler(result)
        return Session(name, handler, request, timeout=timeout, debug=True)

    return maker


@pytest.fixture
def make_manager(app, make_handler, make_session):
    def maker(handler=None):
        if handler is None:
            handler = make_handler([])
        return SessionManager("sm", app, handler, debug=True)

    return maker



import asyncio

from aiohttp import web

import pytest

from aiohttp.test_utils import make_mocked_coro

from sockjs import SessionManager, MSG_OPEN, MSG_CLOSED, MSG_MESSAGE
from sockjs.protocol import FRAME_CLOSE
from sockjs.transports import WebSocketTransport


@pytest.fixture
def make_transport(make_manager, make_request, make_handler, make_fut):
    def maker(method="GET", path="/", query_params={}):
        handler = make_handler(None)
        manager = make_manager(handler)
        request = make_request(method, path, query_params=query_params)
        request.app.freeze()
        session = manager.get("TestSessionWebsocket", create=True, request=request)
        session._wait = make_fut((FRAME_CLOSE, ""))
        return WebSocketTransport(manager, session, request)

    return maker


async def xtest_process_release_acquire_and_remote_closed(make_transport):
    transp = make_transport()
    transp.session.interrupted = False
    transp.manager.acquire = make_mocked_coro()
    transp.manager.release = make_mocked_coro()
    resp = await transp.process()
    await transp.manager.clear()

    assert resp.status == 101
    assert resp.headers.get("upgrade", "").lower() == "websocket"
    assert resp.headers.get("connection", "").lower() == "upgrade"

    transp.session._remote_closed.assert_called_once_with()
    assert transp.manager.acquire.called
    assert transp.manager.release.called


async def test_server_close(app, make_manager, make_request):
    reached_closed = False

    loop = asyncio.get_event_loop()

    async def handler(msg, session):
        nonlocal reached_closed
        if msg.tp == MSG_OPEN:
            asyncio.ensure_future(session._remote_message("TESTMSG"))
            pass

        elif msg.tp == MSG_MESSAGE:
            # To reproduce the ordering which makes the issue
            loop.call_later(0.05, session.close)
        elif msg.tp == MSG_CLOSED:
            reached_closed = True

    app.freeze()

    request = make_request("GET", "/", query_params={})
    manager = SessionManager("sm", app, handler, debug=True)
    session = manager.get("test", create=True)

    transp = WebSocketTransport(manager, session, request)
    await transp.process()

    assert reached_closed is True


async def test_session_has_request(make_transport, make_fut):
    transp = make_transport(method="POST")
    assert isinstance(transp.session.request, web.Request)



from aiohttp import web, hdrs

from ..protocol import loads, ENCODING
from .base import Transport
from .utils import CACHE_CONTROL, session_cookie, cors_headers, cache_headers


class XHRSendTransport(Transport):
    async def process(self):
        request = self.request

        if request.method not in (hdrs.METH_GET, hdrs.METH_POST, hdrs.METH_OPTIONS):
            return web.HTTPForbidden(text="Method is not allowed")

        if self.request.method == hdrs.METH_OPTIONS:
            headers = (
                (hdrs.ACCESS_CONTROL_ALLOW_METHODS, "OPTIONS, POST"),
                (hdrs.CONTENT_TYPE, "application/javascript; charset=UTF-8"),
            )
            headers += session_cookie(request)
            headers += cors_headers(request.headers)
            headers += cache_headers()
            return web.Response(status=204, headers=headers)

        data = await request.read()
        if not data:
            return web.HTTPInternalServerError(text="Payload expected.")

        try:
            messages = loads(data.decode(ENCODING))
        except Exception:
            return web.HTTPInternalServerError(text="Broken JSON encoding.")

        await self.session._remote_messages(messages)

        headers = (
            (hdrs.CONTENT_TYPE, "text/plain; charset=UTF-8"),
            (hdrs.CACHE_CONTROL, CACHE_CONTROL),
        )
        headers += session_cookie(request)
        headers += cors_headers(request.headers)

        return web.Response(status=204, headers=headers)



import asyncio
import logging
from aiohttp import web

import sockjs
from sockjs.transports.eventsource import EventsourceTransport
from sockjs.transports.htmlfile import HTMLFileTransport
from sockjs.transports.xhrstreaming import XHRStreamingTransport


async def echoSession(msg, session):
    if msg.type == sockjs.MSG_MESSAGE:
        session.send(msg.data)


async def closeSessionHander(msg, session):
    if msg.type == sockjs.MSG_OPEN:
        session.close()


async def broadcastSession(msg, session):
    if msg.type == sockjs.MSG_OPEN:
        session.manager.broadcast(msg.data)


if __name__ == '__main__':
    """ Sockjs tests server """
    loop = asyncio.get_event_loop()
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(levelname)s %(message)s')

    HTMLFileTransport.maxsize = 4096
    EventsourceTransport.maxsize = 4096
    XHRStreamingTransport.maxsize = 4096

    app = web.Application(loop=loop)

    sockjs.add_endpoint(
        app, echoSession, name='echo', prefix='/echo')
    sockjs.add_endpoint(
        app, closeSessionHander, name='close', prefix='/close')
    sockjs.add_endpoint(
        app, broadcastSession, name='broadcast', prefix='/broadcast')
    sockjs.add_endpoint(
        app, echoSession, name='wsoff', prefix='/disabled_websocket_echo',
        disable_transports=('websocket',))
    sockjs.add_endpoint(
        app, echoSession, name='cookie', prefix='/cookie_needed_echo',
        cookie_needed=True)

    web.run_app(app)



# pyramid_sockjs

# Session, SessionManager are not imported

from sockjs.session import Session
from sockjs.session import SessionManager
from sockjs.exceptions import SessionIsClosed
from sockjs.exceptions import SessionIsAcquired

from sockjs.protocol import STATE_NEW
from sockjs.protocol import STATE_OPEN
from sockjs.protocol import STATE_CLOSING
from sockjs.protocol import STATE_CLOSED

from sockjs.protocol import MSG_OPEN
from sockjs.protocol import MSG_MESSAGE
from sockjs.protocol import MSG_CLOSE
from sockjs.protocol import MSG_CLOSED

from sockjs.route import get_manager, add_endpoint


__version__ = "0.10.0"


__all__ = (
    "get_manager",
    "add_endpoint",
    "Session",
    "SessionManager",
    "SessionIsClosed",
    "SessionIsAcquired",
    "STATE_NEW",
    "STATE_OPEN",
    "STATE_CLOSING",
    "STATE_CLOSED",
    "MSG_OPEN",
    "MSG_MESSAGE",
    "MSG_CLOSE",
    "MSG_CLOSED",
)



from unittest import mock

from aiohttp import web

import pytest
from aiohttp.test_utils import make_mocked_coro

from sockjs.transports import jsonp


@pytest.fixture
def make_transport(make_request, make_manager, make_handler, make_fut):
    def maker(method="GET", path="/", query_params={}):
        handler = make_handler(None)
        manager = make_manager(handler)
        request = make_request(method, path, query_params=query_params)
        session = manager.get("TestSessionJsonP", create=True, request=request)
        request.app.freeze()
        return jsonp.JSONPolling(manager, session, request)

    return maker


async def test_streaming_send(make_transport):
    trans = make_transport()
    trans.callback = "cb"

    resp = trans.response = mock.Mock()
    resp.write = make_mocked_coro(None)
    stop = await trans.send("text data")
    resp.write.assert_called_with(b'/**/cb("text data");\r\n')
    assert stop


async def test_process(make_transport, make_fut):
    transp = make_transport(query_params={"c": "calback"})
    transp.handle_session = make_fut(1)
    resp = await transp.process()
    assert transp.handle_session.called
    assert resp.status == 200


async def test_process_no_callback(make_transport, make_fut):
    transp = make_transport()
    transp.session = mock.Mock()
    transp.session._remote_closed = make_fut(1)

    resp = await transp.process()
    assert transp.session._remote_closed.called
    assert resp.status == 500


async def test_process_bad_callback(make_transport, make_fut):
    transp = make_transport(query_params={"c": "calback!!!!"})
    transp.session = mock.Mock()
    transp.session._remote_closed = make_fut(1)

    resp = await transp.process()
    assert transp.session._remote_closed.called
    assert resp.status == 500


async def test_process_not_supported(make_transport):
    transp = make_transport(method="PUT")
    resp = await transp.process()
    assert resp.status == 400


async def xtest_process_bad_encoding(make_transport, make_fut):
    transp = make_transport(method="POST")
    transp.request.read = make_fut(b"test")
    transp.request.content_type
    transp.request._content_type = "application/x-www-form-urlencoded"
    resp = await transp.process()
    assert resp.status == 500


async def xtest_process_no_payload(make_transport, make_fut):
    transp = make_transport(method="POST")
    transp.request.read = make_fut(b"d=")
    transp.request.content_type
    transp.request._content_type = "application/x-www-form-urlencoded"
    resp = await transp.process()
    assert resp.status == 500


async def xtest_process_bad_json(make_transport, make_fut):
    transp = make_transport(method="POST")
    transp.request.read = make_fut(b"{]")
    resp = await transp.process()
    assert resp.status == 500


async def xtest_process_message(make_transport, make_fut):
    transp = make_transport(method="POST")
    transp.session._remote_messages = make_fut(1)
    transp.request.read = make_fut(b'["msg1","msg2"]')
    resp = await transp.process()
    assert resp.status == 200
    transp.session._remote_messages.assert_called_with(["msg1", "msg2"])


async def test_session_has_request(make_transport, make_fut):
    transp = make_transport(method="POST")
    transp.session._remote_messages = make_fut(1)
    assert isinstance(transp.session.request, web.Request)



from aiohttp import web

import pytest

from sockjs.transports import xhrsend


@pytest.fixture
def make_transport(make_manager, make_request, make_handler, make_fut):
    def maker(method="GET", path="/", query_params={}):
        handler = make_handler(None)
        manager = make_manager(handler)
        request = make_request(method, path, query_params=query_params)
        request.app.freeze()
        session = manager.get("TestSessionXhrSend", create=True, request=request)
        return xhrsend.XHRSendTransport(manager, session, request)

    return maker


async def test_not_supported_meth(make_transport):
    transp = make_transport(method="PUT")
    resp = await transp.process()
    assert resp.status == 403


async def xtest_no_payload(make_transport, make_fut):
    transp = make_transport()
    transp.request.read = make_fut(b"")
    resp = await transp.process()
    assert resp.status == 500


async def xtest_bad_json(make_transport, make_fut):
    transp = make_transport()
    transp.request.read = make_fut(b"{]")
    resp = await transp.process()
    assert resp.status == 500


async def xtest_post_message(make_transport, make_fut):
    transp = make_transport()
    transp.session._remote_messages = make_fut(1)
    transp.request.read = make_fut(b'["msg1","msg2"]')
    resp = await transp.process()
    assert resp.status == 204
    transp.session._remote_messages.assert_called_with(["msg1", "msg2"])


async def test_OPTIONS(make_transport):
    transp = make_transport(method="OPTIONS")
    resp = await transp.process()
    assert resp.status == 204


async def test_session_has_request(make_transport, make_fut):
    transp = make_transport(method="POST")
    transp.session._remote_messages = make_fut(1)
    assert isinstance(transp.session.request, web.Request)



from unittest import mock
from aiohttp import web

import pytest
from aiohttp.test_utils import make_mocked_coro

from sockjs import protocol
from sockjs.transports import base


@pytest.fixture
def make_transport(make_manager, make_request, make_handler, make_fut):
    def maker(method="GET", path="/", query_params={}):
        handler = make_handler(None)
        manager = make_manager(handler)
        request = make_request(method, path, query_params=query_params)
        request.app.freeze()
        session = manager.get("TestSessionStreaming", create=True, request=request)
        return base.StreamingTransport(manager, session, request)

    return maker


async def test_transport_ctor(make_request):
    manager = object()
    session = object()
    request = make_request("GET", "/")

    transport = base.Transport(manager, session, request)
    assert transport.manager is manager
    assert transport.session is session
    assert transport.request is request


async def test_streaming_send(make_transport):
    trans = make_transport()

    resp = trans.response = mock.Mock()
    resp.write = make_mocked_coro(None)
    stop = await trans.send("text data")
    assert not stop
    assert trans.size == len(b"text data\n")
    resp.write.assert_called_with(b"text data\n")

    trans.maxsize = 1
    stop = await trans.send("text data")
    assert stop


async def test_handle_session_interrupted(make_transport, make_fut):
    trans = make_transport()
    trans.session.interrupted = True
    trans.send = make_fut(1)
    trans.response = web.StreamResponse()
    await trans.handle_session()
    trans.send.assert_called_with('c[1002,"Connection interrupted"]')


async def test_handle_session_closing(make_transport, make_fut):
    trans = make_transport()
    trans.send = make_fut(1)
    trans.session.interrupted = False
    trans.session.state = protocol.STATE_CLOSING
    trans.session._remote_closed = make_fut(1)
    trans.response = web.StreamResponse()
    await trans.handle_session()
    trans.session._remote_closed.assert_called_with()
    trans.send.assert_called_with('c[3000,"Go away!"]')


async def test_handle_session_closed(make_transport, make_fut):
    trans = make_transport()
    trans.send = make_fut(1)
    trans.session.interrupted = False
    trans.session.state = protocol.STATE_CLOSED
    trans.session._remote_closed = make_fut(1)
    trans.response = web.StreamResponse()
    await trans.handle_session()
    trans.session._remote_closed.assert_called_with()
    trans.send.assert_called_with('c[3000,"Go away!"]')


async def test_session_has_request(make_transport, make_fut):
    transp = make_transport(method="POST")
    transp.session._remote_messages = make_fut(1)
    assert isinstance(transp.session.request, web.Request)



# sockjs transports

from .jsonp import JSONPolling
from .websocket import WebSocketTransport
from .htmlfile import HTMLFileTransport
from .eventsource import EventsourceTransport
from .xhr import XHRTransport
from .xhrsend import XHRSendTransport
from .xhrstreaming import XHRStreamingTransport


handlers = {
    "websocket": (True, WebSocketTransport),
    "xhr": (True, XHRTransport),
    "xhr_send": (False, XHRSendTransport),
    "xhr_streaming": (True, XHRStreamingTransport),
    "jsonp": (True, JSONPolling),
    "jsonp_send": (False, JSONPolling),
    "htmlfile": (True, HTMLFileTransport),
    "eventsource": (True, EventsourceTransport),
}



import asyncio
import os
import logging
from aiohttp import web

import sockjs

CHAT_FILE = open(
    os.path.join(os.path.dirname(__file__), 'chat.html'), 'rb').read()


async def chat_msg_handler(msg, session):
    if msg.type == sockjs.MSG_OPEN:
        session.manager.broadcast("Someone joined.")
    elif msg.type == sockjs.MSG_MESSAGE:
        session.manager.broadcast(msg.data)
    elif msg.type == sockjs.MSG_CLOSED:
        session.manager.broadcast("Someone left.")


def index(request):
    return web.Response(body=CHAT_FILE, content_type='text/html')


if __name__ == '__main__':
    """Simple sockjs chat."""
    loop = asyncio.get_event_loop()
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(levelname)s %(message)s')

    app = web.Application(loop=loop)
    app.router.add_route('GET', '/', index)
    sockjs.add_endpoint(app, chat_msg_handler, name='chat', prefix='/sockjs/')

    web.run_app(app)



"""websocket transport"""
import asyncio
from aiohttp import web

from asyncio import ensure_future

from .base import Transport
from ..exceptions import SessionIsClosed
from ..protocol import STATE_CLOSED, FRAME_CLOSE
from ..protocol import loads, close_frame


class WebSocketTransport(Transport):
    async def server(self, ws, session):
        while True:
            try:
                frame, data = await session._wait()
            except SessionIsClosed:
                break
            try:
                await ws.send_str(data)
            except OSError:
                pass  # ignore 'cannot write to closed transport'
            if frame == FRAME_CLOSE:
                try:
                    await ws.close()
                finally:
                    await session._remote_closed()

    async def client(self, ws, session):
        while True:
            msg = await ws.receive()

            if msg.type == web.WSMsgType.text:
                data = msg.data
                if not data:
                    continue

                if data.startswith("["):
                    data = data[1:-1]

                try:
                    text = loads(data)
                except Exception as exc:
                    await session._remote_close(exc)
                    await session._remote_closed()
                    await ws.close(message=b"broken json")
                    break

                await session._remote_message(text)

            elif msg.type == web.WSMsgType.close:
                await session._remote_close()
            elif msg.type in (web.WSMsgType.closed, web.WSMsgType.closing):
                await session._remote_closed()
                break

    async def process(self):
        # start websocket connection
        ws = self.ws = web.WebSocketResponse()
        await ws.prepare(self.request)

        # session was interrupted
        if self.session.interrupted:
            await self.ws.send_str(close_frame(1002, "Connection interrupted"))

        elif self.session.state == STATE_CLOSED:
            await self.ws.send_str(close_frame(3000, "Go away!"))

        else:
            try:
                await self.manager.acquire(self.session)
            except Exception:  # should use specific exception
                await self.ws.send_str(close_frame(3000, "Go away!"))
                await ws.close()
                return ws
            server = ensure_future(self.server(ws, self.session))
            client = ensure_future(self.client(ws, self.session))
            try:
                await asyncio.wait(
                    (server, client), return_when=asyncio.FIRST_COMPLETED
                )
            except asyncio.CancelledError:
                raise
            except Exception as exc:
                await self.session._remote_close(exc)
            finally:
                await self.manager.release(self.session)
                if not server.done():
                    server.cancel()
                if not client.done():
                    client.cancel()

        return ws



"""jsonp transport"""
import re
from urllib.parse import unquote_plus

from aiohttp import web, hdrs

from .base import StreamingTransport
from .utils import CACHE_CONTROL, session_cookie, cors_headers
from ..protocol import dumps, loads, ENCODING


class JSONPolling(StreamingTransport):

    check_callback = re.compile(r"^[a-zA-Z0-9_\.]+$")
    callback = ""

    async def send(self, text):
        data = "/**/%s(%s);\r\n" % (self.callback, dumps(text))
        await self.response.write(data.encode(ENCODING))
        return True

    async def process(self):
        session = self.session
        request = self.request
        meth = request.method

        if request.method == hdrs.METH_GET:
            try:
                callback = self.callback = request.query.get("c")
            except Exception:
                callback = self.callback = request.GET.get("c")

            if not callback:
                await self.session._remote_closed()
                return web.HTTPInternalServerError(text='"callback" parameter required')

            elif not self.check_callback.match(callback):
                await self.session._remote_closed()
                return web.HTTPInternalServerError(text='invalid "callback" parameter')

            headers = (
                (hdrs.CONTENT_TYPE, "application/javascript; charset=UTF-8"),
                (hdrs.CACHE_CONTROL, CACHE_CONTROL),
            )
            headers += session_cookie(request)
            headers += cors_headers(request.headers)

            resp = self.response = web.StreamResponse(headers=headers)
            await resp.prepare(request)

            await self.handle_session()
            return resp

        elif request.method == hdrs.METH_POST:
            data = await request.read()

            ctype = request.content_type.lower()
            if ctype == "application/x-www-form-urlencoded":
                if not data.startswith(b"d="):
                    return web.HTTPInternalServerError(text="Payload expected.")

                data = unquote_plus(data[2:].decode(ENCODING))
            else:
                data = data.decode(ENCODING)

            if not data:
                return web.HTTPInternalServerError(text="Payload expected.")

            try:
                messages = loads(data)
            except Exception:
                return web.HTTPInternalServerError(text="Broken JSON encoding.")

            await session._remote_messages(messages)

            headers = (
                (hdrs.CONTENT_TYPE, "text/html;charset=UTF-8"),
                (hdrs.CACHE_CONTROL, CACHE_CONTROL),
            )
            headers += session_cookie(request)
            return web.Response(body=b"ok", headers=headers)

        else:
            return web.HTTPBadRequest(text="No support for such method: %s" % meth)



from unittest import mock
from asyncio import Future

import pytest

from sockjs.exceptions import SessionIsClosed
from sockjs.protocol import FRAME_CLOSE, FRAME_HEARTBEAT
from sockjs.transports.rawwebsocket import RawWebSocketTransport

from aiohttp import WSMessage, WSMsgType


@pytest.fixture
def make_transport(make_request, make_fut):
    def maker(method="GET", path="/", query_params={}):
        manager = mock.Mock()
        session = mock.Mock()
        session._remote_closed = make_fut(1)
        session._wait = make_fut((FRAME_CLOSE, ""))
        request = make_request(method, path, query_params=query_params)
        request.app.freeze()
        return RawWebSocketTransport(manager, session, request)

    return maker


async def xtest_ticks_pong(make_transport, make_fut):
    transp = make_transport()

    pong = WSMessage(type=WSMsgType.PONG, data=b"", extra="")
    close = WSMessage(type=WSMsgType.closing, data=b"", extra="")

    future = Future()
    future.set_result(pong)

    future2 = Future()
    future2.set_result(close)

    ws = mock.Mock()
    ws.receive.side_effect = [future, future2]

    session = transp.session

    await transp.client(ws, session)
    assert session._tick.called


async def test_sends_ping(make_transport, make_fut):
    transp = make_transport()

    future = Future()
    future.set_result(False)

    ws = mock.Mock()
    ws.ping.side_effect = [future]

    hb_future = Future()
    hb_future.set_result((FRAME_HEARTBEAT, b""))

    session_close_future = Future()
    session_close_future.set_exception(SessionIsClosed)

    session = mock.Mock()
    session._wait.side_effect = [hb_future, session_close_future]

    await transp.server(ws, session)
    assert ws.ping.called



"""Exceptions"""


class SockjsException(Exception):
    """Base sockjs exception."""


class SessionIsAcquired(SockjsException):
    """Session is acquired."""


class SessionIsClosed(SockjsException):
    """Session is closed."""



import http.cookies
from aiohttp import hdrs
from datetime import datetime, timedelta


CACHE_CONTROL = "no-store, no-cache, no-transform, must-revalidate, max-age=0"


def cors_headers(headers, nocreds=False):
    origin = headers.get(hdrs.ORIGIN, "*")
    cors = ((hdrs.ACCESS_CONTROL_ALLOW_ORIGIN, origin),)

    ac_headers = headers.get(hdrs.ACCESS_CONTROL_REQUEST_HEADERS)
    if ac_headers:
        cors += ((hdrs.ACCESS_CONTROL_ALLOW_HEADERS, ac_headers),)

    if origin != "*":
        return cors + ((hdrs.ACCESS_CONTROL_ALLOW_CREDENTIALS, "true"),)
    else:
        return cors


def session_cookie(request):
    cookie = request.cookies.get("JSESSIONID", "dummy")
    cookies = http.cookies.SimpleCookie()
    cookies["JSESSIONID"] = cookie
    cookies["JSESSIONID"]["path"] = "/"
    return ((hdrs.SET_COOKIE, cookies["JSESSIONID"].output(header="")[1:]),)


td365 = timedelta(days=365)
td365seconds = str(
    (td365.microseconds + (td365.seconds + td365.days * 24 * 3600) * 10 ** 6) // 10 ** 6
)


def cache_headers():
    d = datetime.now() + td365
    return (
        (hdrs.ACCESS_CONTROL_MAX_AGE, td365seconds),
        (hdrs.CACHE_CONTROL, "max-age=%s, public" % td365seconds),
        (hdrs.EXPIRES, d.strftime("%a, %d %b %Y %H:%M:%S")),
    )



from aiohttp import web, hdrs

from .base import StreamingTransport
from .utils import CACHE_CONTROL, session_cookie, cors_headers, cache_headers


class XHRStreamingTransport(StreamingTransport):

    maxsize = 131072  # 128K bytes
    open_seq = b"h" * 2048 + b"\n"

    async def process(self):
        request = self.request
        headers = (
            (hdrs.CONNECTION, request.headers.get(hdrs.CONNECTION, "close")),
            (hdrs.CONTENT_TYPE, "application/javascript; charset=UTF-8"),
            (hdrs.CACHE_CONTROL, CACHE_CONTROL),
        )

        headers += session_cookie(request)
        headers += cors_headers(request.headers)

        if request.method == hdrs.METH_OPTIONS:
            headers += ((hdrs.ACCESS_CONTROL_ALLOW_METHODS, "OPTIONS, POST"),)
            headers += cache_headers()
            return web.Response(status=204, headers=headers)

        # open sequence (sockjs protocol)
        resp = self.response = web.StreamResponse(headers=headers)
        resp.force_close()
        await resp.prepare(request)
        await resp.write(self.open_seq)

        # event loop
        await self.handle_session()

        return resp



import aiohttp
import asyncio

from ..exceptions import SessionIsAcquired, SessionIsClosed
from ..protocol import close_frame, ENCODING
from ..protocol import STATE_CLOSING, STATE_CLOSED, FRAME_CLOSE, FRAME_MESSAGE


class Transport:
    def __init__(self, manager, session, request):
        self.manager = manager
        self.session = session
        self.request = request


class StreamingTransport(Transport):

    timeout = None
    maxsize = 131072  # 128K bytes

    def __init__(self, manager, session, request):
        super().__init__(manager, session, request)

        self.size = 0
        self.response = None

    async def send(self, text):
        blob = (text + "\n").encode(ENCODING)
        await self.response.write(blob)

        self.size += len(blob)
        if self.size > self.maxsize:
            return True
        else:
            return False

    async def handle_session(self):
        assert self.response is not None, "Response is not specified."

        # session was interrupted
        if self.session.interrupted:
            await self.send(close_frame(1002, "Connection interrupted"))

        # session is closing or closed
        elif self.session.state in (STATE_CLOSING, STATE_CLOSED):
            await self.session._remote_closed()
            await self.send(close_frame(3000, "Go away!"))

        else:
            # acquire session
            try:
                await self.manager.acquire(self.session)
            except SessionIsAcquired:
                await self.send(close_frame(2010, "Another connection still open"))
            else:
                try:
                    while True:
                        if self.timeout:
                            try:
                                frame, text = await asyncio.wait_for(
                                    self.session._wait(), timeout=self.timeout
                                )
                            except asyncio.futures.TimeoutError:
                                frame, text = FRAME_MESSAGE, "a[]"
                        else:
                            frame, text = await self.session._wait()

                        if frame == FRAME_CLOSE:
                            await self.session._remote_closed()
                            await self.send(text)
                            return
                        else:
                            stop = await self.send(text)
                            if stop:
                                break
                except asyncio.CancelledError:
                    await self.session._remote_close(exc=aiohttp.ClientConnectionError)
                    await self.session._remote_closed()
                    raise
                except SessionIsClosed:
                    pass
                finally:
                    await self.manager.release(self.session)



import asyncio

from aiohttp import web
from multidict import CIMultiDict

from sockjs import protocol


async def test_info(make_route, make_request):
    route = make_route()
    request = make_request("GET", "/sm/")

    response = await route.info(request)
    info = protocol.loads(response.body.decode("utf-8"))

    assert info["websocket"]
    assert info["cookie_needed"]


async def test_info_entropy(make_route, make_request):
    route = make_route()
    request = make_request("GET", "/sm/")

    response = await route.info(request)
    entropy1 = protocol.loads(response.body.decode("utf-8"))["entropy"]

    response = await route.info(request)
    entropy2 = protocol.loads(response.body.decode("utf-8"))["entropy"]

    assert entropy1 != entropy2


async def test_info_options(make_route, make_request):
    route = make_route()
    request = make_request("OPTIONS", "/sm/")
    response = await route.info_options(request)

    assert response.status == 204

    headers = response.headers
    assert "Access-Control-Max-Age" in headers
    assert "Cache-Control" in headers
    assert "Expires" in headers
    assert "Set-Cookie" in headers
    assert "access-control-allow-credentials" in headers
    assert "access-control-allow-origin" in headers


async def test_greeting(make_route, make_request):
    route = make_route()
    request = make_request("GET", "/sm/")
    response = await route.greeting(request)

    assert response.body == b"Welcome to SockJS!\n"


async def test_iframe(make_route, make_request):
    route = make_route()
    request = make_request("GET", "/sm/")

    response = await route.iframe(request)
    text = """<!DOCTYPE html>
<html>
<head>
<meta http-equiv="X-UA-Compatible" content="IE=edge" />
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
  <script src="http:sockjs-cdn"></script>
  <script>
    document.domain = document.domain;
    SockJS.bootstrap_iframe();
  </script>
</head>
<body>
  <h2>Don't panic!</h2>
  <p>This is a SockJS hidden iframe. It's used for cross domain magic.</p>
</body>
</html>"""

    assert response.body.decode("utf-8") == text
    assert "ETag" in response.headers


async def test_iframe_cache(make_route, make_request):
    route = make_route()
    request = make_request(
        "GET", "/sm/", headers=CIMultiDict({"IF-NONE-MATCH": "test"})
    )
    response = await route.iframe(request)

    assert response.status == 304


async def test_handler_unknown_transport(make_route, make_request):
    route = make_route()
    request = make_request("GET", "/sm/", match_info={"transport": "unknown"})

    res = await route.handler(request)
    assert isinstance(res, web.HTTPNotFound)


async def test_handler_emptry_session(make_route, make_request):
    route = make_route()
    request = make_request(
        "GET", "/sm/", match_info={"transport": "websocket", "session": ""}
    )
    res = await route.handler(request)
    assert isinstance(res, web.HTTPNotFound)


async def test_handler_bad_session_id(make_route, make_request):
    route = make_route()
    request = make_request(
        "GET",
        "/sm/",
        match_info={"transport": "websocket", "session": "test.1", "server": "000"},
    )
    res = await route.handler(request)
    assert isinstance(res, web.HTTPNotFound)


async def test_handler_bad_server_id(make_route, make_request):
    route = make_route()
    request = make_request(
        "GET",
        "/sm/",
        match_info={"transport": "websocket", "session": "test", "server": "test.1"},
    )
    res = await route.handler(request)
    assert isinstance(res, web.HTTPNotFound)


async def test_new_session_before_read(make_route, make_request):
    route = make_route()
    request = make_request(
        "GET",
        "/sm/",
        match_info={"transport": "xhr_send", "session": "s1", "server": "000"},
    )
    res = await route.handler(request)
    assert isinstance(res, web.HTTPNotFound)


async def _test_transport(make_route, make_request):
    route = make_route()
    request = make_request(
        "GET", "/sm/", match_info={"transport": "xhr", "session": "s1", "server": "000"}
    )

    params = []

    class Transport:
        def __init__(self, manager, session, request):
            params.append((manager, session, request))

        def process(self):
            return web.HTTPOk()

    route = make_route(handlers={"test": (True, Transport)})
    res = await route.handler(request)
    assert isinstance(res, web.HTTPOk)
    assert params[0] == (route.manager, route.manager["s1"], request)


async def test_fail_transport(make_route, make_request):
    request = make_request(
        "GET",
        "/sm/",
        match_info={"transport": "test", "session": "session", "server": "000"},
    )

    params = []

    class Transport:
        def __init__(self, manager, session, request):
            params.append((manager, session, request))

        def process(self):
            raise Exception("Error")

    route = make_route(handlers={"test": (True, Transport)})
    res = await route.handler(request)
    assert isinstance(res, web.HTTPInternalServerError)


async def test_release_session_for_failed_transport(make_route, make_request):
    request = make_request(
        "GET",
        "/sm/",
        match_info={"transport": "test", "session": "s1", "server": "000"},
    )

    class Transport:
        def __init__(self, manager, session, request):
            self.manager = manager
            self.session = session

        async def process(self):
            await self.manager.acquire(self.session)
            raise Exception("Error")

    route = make_route(handlers={"test": (True, Transport)})
    res = await route.handler(request)
    assert isinstance(res, web.HTTPInternalServerError)

    s1 = route.manager["s1"]
    assert not route.manager.is_acquired(s1)


async def test_raw_websocket(make_route, make_request, mocker):
    ws = mocker.patch("sockjs.route.RawWebSocketTransport")
    loop = asyncio.get_event_loop()
    ws.return_value.process.return_value = loop.create_future()
    ws.return_value.process.return_value.set_result(web.HTTPOk())

    route = make_route()
    request = make_request("GET", "/sm/", headers=CIMultiDict({}))
    res = await route.websocket(request)

    assert isinstance(res, web.HTTPOk)
    assert ws.called
    assert ws.return_value.process.called


async def _test_raw_websocket_fail(make_route, make_request):
    route = make_route()
    request = make_request("GET", "/sm/")
    res = await route.websocket(request)
    assert not isinstance(res, web.HTTPNotFound)



from aiohttp import web, hdrs

from .base import StreamingTransport
from .utils import CACHE_CONTROL, session_cookie, cors_headers, cache_headers


class XHRTransport(StreamingTransport):
    """Long polling derivative transports,
    used for XHRPolling and JSONPolling."""

    maxsize = 0

    async def process(self):
        request = self.request

        if request.method == hdrs.METH_OPTIONS:
            headers = (
                (hdrs.CONTENT_TYPE, "application/javascript; charset=UTF-8"),
                (hdrs.ACCESS_CONTROL_ALLOW_METHODS, "OPTIONS, POST"),
            )
            headers += session_cookie(request)
            headers += cors_headers(request.headers)
            headers += cache_headers()
            return web.Response(status=204, headers=headers)

        headers = (
            (hdrs.CONTENT_TYPE, "application/javascript; charset=UTF-8"),
            (hdrs.CACHE_CONTROL, CACHE_CONTROL),
        )
        headers += session_cookie(request)
        headers += cors_headers(request.headers)

        resp = self.response = web.StreamResponse(headers=headers)
        await resp.prepare(request)

        await self.handle_session()
        return resp



import codecs
import os
import re
from setuptools import setup, find_packages

with codecs.open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "sockjs", "__init__.py"),
    "r",
    "latin1",
) as fp:
    try:
        version = re.findall(r'^__version__ = "([^"]+)"\r?$', fp.read(), re.M)[0]
    except IndexError:
        raise RuntimeError("Unable to determine version.")

install_requires = ["aiohttp >= 3.0.0"]


def read(f):
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()


setup(
    name="sockjs",
    version=version,
    description=("SockJS server implementation for aiohttp."),
    long_description="\n\n".join((read("README.rst"), read("CHANGES.txt"))),
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Internet :: WWW/HTTP",
        "Framework :: AsyncIO",
    ],
    author="Nikolay Kim",
    author_email="fafhrd91@gmail.com",
    url="https://github.com/aio-libs/sockjs/",
    license="Apache 2",
    packages=find_packages(),
    python_requires=">=3.5.3",
    install_requires=install_requires,
    include_package_data=True,
    zip_safe=False,
)



import warnings
import asyncio
import collections
import logging
from datetime import datetime, timedelta

from asyncio import ensure_future

from .protocol import STATE_NEW, STATE_OPEN, STATE_CLOSING, STATE_CLOSED
from .protocol import FRAME_OPEN, FRAME_CLOSE
from .protocol import FRAME_MESSAGE, FRAME_MESSAGE_BLOB, FRAME_HEARTBEAT
from .exceptions import SessionIsAcquired, SessionIsClosed

from .protocol import MSG_CLOSE, MSG_MESSAGE
from .protocol import close_frame, message_frame, messages_frame
from .protocol import SockjsMessage, OpenMessage, ClosedMessage


log = logging.getLogger("sockjs")


class Session(object):
    """ SockJS session object

    ``state``: Session state

    ``manager``: Session manager that hold this session

    ``acquired``: Acquired state, indicates that transport is using session

    ``timeout``: Session timeout

    """

    manager = None
    acquired = False
    state = STATE_NEW
    interrupted = False
    exception = None

    def __init__(
        self, id, handler, request, *, timeout=timedelta(seconds=10), debug=False
    ):
        self.id = id
        self.handler = handler
        self.request = request
        self.expired = False
        self.timeout = timeout
        self.expires = datetime.now() + timeout

        self._hits = 0
        self._heartbeats = 0
        self._heartbeat_transport = False
        self._debug = debug
        self._waiter = None
        self._queue = collections.deque()

    def __str__(self):
        result = ["id=%r" % (self.id,)]

        if self.state == STATE_OPEN:
            result.append("connected")
        elif self.state == STATE_CLOSED:
            result.append("closed")
        else:
            result.append("disconnected")

        if self.acquired:
            result.append("acquired")

        if len(self._queue):
            result.append("queue[%s]" % len(self._queue))
        if self._hits:
            result.append("hits=%s" % self._hits)
        if self._heartbeats:
            result.append("heartbeats=%s" % self._heartbeats)

        return " ".join(result)

    def _tick(self, timeout=None):
        if timeout is None:
            self.expires = datetime.now() + self.timeout
        else:
            self.expires = datetime.now() + timeout

    async def _acquire(self, manager, heartbeat=True):
        self.acquired = True
        self.manager = manager
        self._heartbeat_transport = heartbeat

        self._tick()
        self._hits += 1

        if self.state == STATE_NEW:
            log.debug("open session: %s", self.id)
            self.state = STATE_OPEN
            self._feed(FRAME_OPEN, FRAME_OPEN)
            try:
                await self.handler(OpenMessage, self)
            except asyncio.CancelledError:
                raise
            except Exception as exc:
                self.state = STATE_CLOSING
                self.exception = exc
                self.interrupted = True
                self._feed(FRAME_CLOSE, (3000, "Internal error"))
                log.exception("Exception in open session handling.")

    def _release(self):
        self.acquired = False
        self.manager = None
        self._heartbeat_transport = False

    def _heartbeat(self):
        self._heartbeats += 1
        if self._heartbeat:
            self._feed(FRAME_HEARTBEAT, FRAME_HEARTBEAT)

    def _feed(self, frame, data):
        # pack messages
        if frame == FRAME_MESSAGE:
            if self._queue and self._queue[-1][0] == FRAME_MESSAGE:
                self._queue[-1][1].append(data)
            else:
                self._queue.append((frame, [data]))
        else:
            self._queue.append((frame, data))

        # notify waiter
        waiter = self._waiter
        if waiter is not None:
            self._waiter = None
            if not waiter.cancelled():
                waiter.set_result(True)

    async def _wait(self, pack=True):
        if not self._queue and self.state != STATE_CLOSED:
            assert not self._waiter
            loop = asyncio.get_event_loop()
            self._waiter = loop.create_future()
            await self._waiter

        if self._queue:
            frame, payload = self._queue.popleft()
            if pack:
                if frame == FRAME_CLOSE:
                    return FRAME_CLOSE, close_frame(*payload)
                elif frame == FRAME_MESSAGE:
                    return FRAME_MESSAGE, messages_frame(payload)

            return frame, payload
        else:
            raise SessionIsClosed()

    async def _remote_close(self, exc=None):
        """close session from remote."""
        if self.state in (STATE_CLOSING, STATE_CLOSED):
            return

        log.info("close session: %s", self.id)
        self.state = STATE_CLOSING
        if exc is not None:
            self.exception = exc
            self.interrupted = True
        try:
            await self.handler(SockjsMessage(MSG_CLOSE, exc), self)
        except Exception:
            log.exception("Exception in close handler.")

    async def _remote_closed(self):
        if self.state == STATE_CLOSED:
            return

        log.info("session closed: %s", self.id)
        self.state = STATE_CLOSED
        self.expire()
        try:
            await self.handler(ClosedMessage, self)
        except Exception:
            log.exception("Exception in closed handler.")

        # notify waiter
        waiter = self._waiter
        if waiter is not None:
            self._waiter = None
            if not waiter.cancelled():
                waiter.set_result(True)

    async def _remote_message(self, msg):
        log.debug("incoming message: %s, %s", self.id, msg[:200])
        self._tick()

        try:
            await self.handler(SockjsMessage(MSG_MESSAGE, msg), self)
        except Exception:
            log.exception("Exception in message handler.")

    async def _remote_messages(self, messages):
        self._tick()

        for msg in messages:
            log.debug("incoming message: %s, %s", self.id, msg[:200])
            try:
                await self.handler(SockjsMessage(MSG_MESSAGE, msg), self)
            except Exception:
                log.exception("Exception in message handler.")

    def expire(self):
        """Manually expire a session."""
        self.expired = True

    def send(self, msg):
        """send message to client."""
        assert isinstance(msg, str), "String is required"

        if self._debug:
            log.info("outgoing message: %s, %s", self.id, str(msg)[:200])

        if self.state != STATE_OPEN:
            return

        self._feed(FRAME_MESSAGE, msg)

    def send_frame(self, frm):
        """send message frame to client."""
        if self._debug:
            log.info("outgoing message: %s, %s", self.id, frm[:200])

        if self.state != STATE_OPEN:
            return

        self._feed(FRAME_MESSAGE_BLOB, frm)

    def close(self, code=3000, reason="Go away!"):
        """close session"""
        if self.state in (STATE_CLOSING, STATE_CLOSED):
            return

        if self._debug:
            log.debug("close session: %s", self.id)

        self.state = STATE_CLOSING
        self._feed(FRAME_CLOSE, (code, reason))


_marker = object()


class SessionManager(dict):
    """A basic session manager."""

    _hb_handle = None  # heartbeat event loop timer
    _hb_task = None  # gc task

    def __init__(
        self,
        name,
        app,
        handler,
        heartbeat=25.0,
        timeout=timedelta(seconds=5),
        debug=False,
    ):
        self.name = name
        self.route_name = "sockjs-url-%s" % name
        self.app = app
        self.handler = handler
        self.factory = Session
        self.acquired = {}
        self.sessions = []
        self.heartbeat = heartbeat
        self.timeout = timeout
        self.debug = debug

    def route_url(self, request):
        return request.route_url(self.route_name)

    @property
    def started(self):
        return self._hb_handle is not None

    def start(self):
        if not self._hb_handle:
            loop = asyncio.get_event_loop()
            self._hb_handle = loop.call_later(self.heartbeat, self._heartbeat)

    def stop(self):
        if self._hb_handle is not None:
            self._hb_handle.cancel()
            self._hb_handle = None
        if self._hb_task is not None:
            self._hb_task.cancel()
            self._hb_task = None

    def _heartbeat(self):
        if self._hb_task is None:
            self._hb_task = ensure_future(self._heartbeat_task())

    async def _heartbeat_task(self):
        sessions = self.sessions

        if sessions:
            now = datetime.now()

            idx = 0
            while idx < len(sessions):
                session = sessions[idx]

                session._heartbeat()

                if session.expires < now:
                    # Session is to be GC'd immedietely
                    if session.id in self.acquired:
                        await self.release(session)
                    if session.state == STATE_OPEN:
                        await session._remote_close()
                    if session.state == STATE_CLOSING:
                        await session._remote_closed()

                    del self[session.id]
                    del self.sessions[idx]
                    continue

                idx += 1

        self._hb_task = None
        loop = asyncio.get_event_loop()
        self._hb_handle = loop.call_later(self.heartbeat, self._heartbeat)

    def _add(self, session):
        if session.expired:
            raise ValueError("Can not add expired session")

        session.manager = self
        session.registry = self.app

        self[session.id] = session
        self.sessions.append(session)
        return session

    def get(self, id, create=False, request=None, default=_marker):
        session = super(SessionManager, self).get(id, None)
        if session is None:
            if create:
                session = self._add(
                    self.factory(
                        id,
                        self.handler,
                        request,
                        timeout=self.timeout,
                        debug=self.debug,
                    )
                )
            else:
                if default is not _marker:
                    return default
                raise KeyError(id)

        return session

    async def acquire(self, s):
        sid = s.id

        if sid in self.acquired:
            raise SessionIsAcquired("Another connection still open")
        if sid not in self:
            raise KeyError("Unknown session")

        await s._acquire(self)

        self.acquired[sid] = True
        return s

    def is_acquired(self, session):
        return session.id in self.acquired

    async def release(self, s):
        if s.id in self.acquired:
            s._release()
            del self.acquired[s.id]

    def active_sessions(self):
        for session in self.values():
            if not session.expired:
                yield session

    async def clear(self):
        """Manually expire all sessions in the pool."""
        for session in list(self.values()):
            if session.state != STATE_CLOSED:
                await session._remote_closed()

        self.sessions.clear()
        super(SessionManager, self).clear()

    def broadcast(self, message):
        blob = message_frame(message)

        for session in self.values():
            if not session.expired:
                session.send_frame(blob)

    def __del__(self):
        if len(self.sessions):
            warnings.warn(
                "Unclosed sessions! "
                "Please call `await SessionManager.clear()` before del",
                RuntimeWarning,
            )
        self.stop()



""" iframe-htmlfile transport """
import re
from aiohttp import web, hdrs

from ..protocol import dumps, ENCODING
from .base import StreamingTransport
from .utils import CACHE_CONTROL, session_cookie, cors_headers


PRELUDE1 = b"""
<!doctype html>
<html><head>
  <meta http-equiv="X-UA-Compatible" content="IE=edge" />
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
</head><body><h2>Don't panic!</h2>
  <script>
    document.domain = document.domain;
    var c = parent."""

PRELUDE2 = b""";
    c.start();
    function p(d) {c.message(d);};
    window.onload = function() {c.stop();};
  </script>"""


class HTMLFileTransport(StreamingTransport):

    maxsize = 131072  # 128K bytes
    check_callback = re.compile(r"^[a-zA-Z0-9_\.]+$")

    async def send(self, text):
        blob = ("<script>\np(%s);\n</script>\r\n" % dumps(text)).encode(ENCODING)
        await self.response.write(blob)

        self.size += len(blob)
        if self.size > self.maxsize:
            return True
        else:
            return False

    async def process(self):
        request = self.request

        try:
            callback = request.query.get("c", None)
        except Exception:
            callback = request.GET.get("c", None)

        if callback is None:
            await self.session._remote_closed()
            return web.HTTPInternalServerError(text='"callback" parameter required')

        elif not self.check_callback.match(callback):
            await self.session._remote_closed()
            return web.HTTPInternalServerError(text='invalid "callback" parameter')

        headers = (
            (hdrs.CONTENT_TYPE, "text/html; charset=UTF-8"),
            (hdrs.CACHE_CONTROL, CACHE_CONTROL),
            (hdrs.CONNECTION, "close"),
        )
        headers += session_cookie(request)
        headers += cors_headers(request.headers)

        # open sequence (sockjs protocol)
        resp = self.response = web.StreamResponse(headers=headers)
        await resp.prepare(self.request)
        await resp.write(
            b"".join((PRELUDE1, callback.encode("utf-8"), PRELUDE2, b" " * 1024))
        )

        # handle session
        await self.handle_session()

        return resp



"""raw websocket transport."""
import asyncio
from aiohttp import web

from asyncio import ensure_future

from .base import Transport
from ..exceptions import SessionIsClosed
from ..protocol import FRAME_CLOSE, FRAME_MESSAGE, FRAME_MESSAGE_BLOB, FRAME_HEARTBEAT


class RawWebSocketTransport(Transport):
    async def server(self, ws, session):
        while True:
            try:
                frame, data = await session._wait(pack=False)
            except SessionIsClosed:
                break

            if frame == FRAME_MESSAGE:
                for text in data:
                    await ws.send_str(text)
            elif frame == FRAME_MESSAGE_BLOB:
                data = data[1:]
                if data.startswith("["):
                    data = data[1:-1]
                await ws.send_str(data)
            elif frame == FRAME_HEARTBEAT:
                await ws.ping()
            elif frame == FRAME_CLOSE:
                try:
                    await ws.close(message="Go away!")
                finally:
                    await session._remote_closed()

    async def client(self, ws, session):
        while True:
            msg = await ws.receive()

            if msg.type == web.WSMsgType.text:
                if not msg.data:
                    continue
                await self.session._remote_message(msg.data)
            elif msg.type == web.WSMsgType.close:
                await self.session._remote_close()
            elif msg.type in (web.WSMsgType.closed, web.WSMsgType.closing):
                await self.session._remote_closed()
                break
            elif msg.type == web.WSMsgType.PONG:
                self.session._tick()

    async def process(self):
        # start websocket connection
        ws = self.ws = web.WebSocketResponse(autoping=False)
        await ws.prepare(self.request)

        try:
            await self.manager.acquire(self.session)
        except Exception:  # should use specific exception
            await ws.close(message="Go away!")
            return ws

        server = ensure_future(self.server(ws, self.session))
        client = ensure_future(self.client(ws, self.session))
        try:
            await asyncio.wait((server, client), return_when=asyncio.FIRST_COMPLETED)
        except asyncio.CancelledError:
            raise
        except Exception as exc:
            await self.session._remote_close(exc)
        finally:
            await self.manager.release(self.session)
            if not server.done():
                server.cancel()
            if not client.done():
                client.cancel()

        return ws



from unittest import mock
from aiohttp import web

import pytest
from aiohttp.test_utils import make_mocked_coro

from sockjs.transports import htmlfile


@pytest.fixture
def make_transport(make_manager, make_request, make_handler, make_fut):
    def maker(method="GET", path="/", query_params={}):
        handler = make_handler(None)
        manager = make_manager(handler)
        request = make_request(method, path, query_params=query_params)
        request.app.freeze()
        session = manager.get("TestSessionHtmlFile", create=True, request=request)
        return htmlfile.HTMLFileTransport(manager, session, request)

    return maker


async def test_streaming_send(make_transport):
    trans = make_transport()

    resp = trans.response = mock.Mock()
    resp.write = make_mocked_coro(None)
    stop = await trans.send("text data")
    resp.write.assert_called_with(b'<script>\np("text data");\n</script>\r\n')
    assert not stop
    assert trans.size == len(b'<script>\np("text data");\n</script>\r\n')

    trans.maxsize = 1
    stop = trans.send("text data")
    assert stop


async def test_process(make_transport, make_fut):
    transp = make_transport(query_params={"c": "calback"})
    transp.handle_session = make_fut(1)
    resp = await transp.process()
    assert transp.handle_session.called
    assert resp.status == 200


async def test_process_no_callback(make_transport, make_fut):
    transp = make_transport()
    transp.session = mock.Mock()
    transp.session._remote_closed = make_fut(1)

    resp = await transp.process()
    assert transp.session._remote_closed.called
    assert resp.status == 500


async def test_process_bad_callback(make_transport, make_fut):
    transp = make_transport(query_params={"c": "calback!!!!"})
    transp.session = mock.Mock()
    transp.session._remote_closed = make_fut(1)

    resp = await transp.process()
    assert transp.session._remote_closed.called
    assert resp.status == 500


async def test_session_has_request(make_transport, make_fut):
    transp = make_transport(method="POST")
    transp.session._remote_messages = make_fut(1)
    assert isinstance(transp.session.request, web.Request)



import json

from sockjs import protocol


def test_encode():
    assert protocol.dumps({}) == json.dumps({})
    assert protocol.dumps(["test"]) == json.dumps(["test"])
    assert protocol.dumps('"test"') == json.dumps('"test"')


def test_decode():
    assert protocol.loads(json.dumps({})) == {}
    assert protocol.loads(json.dumps(["test"])) == ["test"]
    assert protocol.loads(json.dumps('"test"')) == '"test"'


def test_close_frame():
    msg = protocol.close_frame(1000, "Internal error")
    assert msg == 'c[1000,"Internal error"]'


def test_message_frame():
    msg = protocol.message_frame("msg1")
    assert msg == "a%s" % protocol.dumps(["msg1"])


def test_messages_frame():
    msg = protocol.messages_frame(["msg1", "msg2"])
    assert msg == "a%s" % protocol.dumps(["msg1", "msg2"])
